import json
from typing import Any, Dict, List, Union, Tuple, Optional

from .base import ConfItem, IConf, ConfTypes, logger
from .parser import ValueParser


class ConfDB(IConf):
    """
    基于数据加的配置存储
    """
    SQL = {
        # 查询表是否存在
        'table_exists': """SELECT * FROM `information_schema`.`TABLES`
            WHERE `TABLE_SCHEMA`='{db_name}' AND `TABLE_NAME`='{table_name}' LIMIT 1""",
        # 创建表结构
        'create_table': """CREATE TABLE `{table_name}` (
        `id` VARCHAR(128) PRIMARY KEY NOT NULL,
        `name` VARCHAR(64),
        `type` VARCHAR(32) DEFAULT '',
        `value` TEXT,
        `user_value` TEXT,
        `value_type` VARCHAR(32),
        `enabled` BIT DEFAULT 1,
        `desc` VARCHAR(256)
        ) ENGINE=InnoDB DEFAULT CHARSET=utf8""",
        # 加载所有数据
        'load_all': 'SELECT * FROM `{table_name}`',
        # 通过 id 查询一项数据
        'get_by_id': 'SELECT * FROM `{table_name}` WHERE `id`=%s',
        # 加载匹配指定前缀的数据
        'match_items': 'SELECT * FROM `{table_name}` WHERE `id` LIKE %s',
        # 更新数据值
        'update_value_by_id': 'UPDATE `{table_name}` SET `{value_field}`=%s, `value_type`=%s WHERE `id`=%s',
        # 更新所有字段
        'update_by_id': 'UPDATE `{table_name}` SET '
                        '`name`=%s,'
                        '`value`=%s,'
                        '`user_value`=%s,'
                        '`value_type`=%s,'
                        '`type`=%s,'
                        '`desc`=%s,'
                        '`enabled`=%s'
                        ' WHERE `id`=%s',
        # 添加项
        'add_item': 'INSERT INTO `{table_name}`'
                    '(`id`,`name`,`value`,`user_value`,`value_type`,`type`,`desc`,`enabled`)'
                    'VALUES(%s,%s,%s,%s,%s,%s,%s,%s)',
        # 移除项
        'remove_item': 'DELETE FROM `{table_name}` WHERE `id`=%s',
        # 清空数据
        'clear_table': 'TRUNCATE TABLE `{table_name}`'
    }

    def __init__(self,
                 database: str,
                 host='127.0.0.1',
                 port=3306,
                 user='root',
                 password='',
                 creator=None,
                 parser: ValueParser = None,
                 table_name='conf_items', **extra_pool_option):
        from dbutils.pooled_db import PooledDB

        self.table_name = table_name
        if creator is None:
            import pymysql
            creator = pymysql
        self.pool_option = {
            'creator': creator,
            'database': database,
            'host': host,
            'port': port,
            'user': user,
            'password': password,
            **extra_pool_option
        }
        logger.debug(
            'Initialize database configuration with table "{table_name}" via URI '
            'mysql://{host}:{port}/{database}'.format(
                table_name=table_name,
                **self.pool_option
            )
        )
        self.pool = PooledDB(**self.pool_option)
        self._tested = False
        super(ConfDB, self).__init__(parser)

    def _execute(self, sql: str, *args):
        if not self._tested:
            self._tested = True
            if not self._table_exists():
                self._create_table()

        import pymysql
        conn = self.pool.connection(True)
        # try:
        #     print('[ConfDB] ' + (sql % args if args else sql))
        # except:
        #     print('[ConfDB] ' + sql)

        cursor = None
        is_query = sql.startswith('SELECT')
        try:
            cursor = conn.cursor(pymysql.cursors.DictCursor)
            rows = cursor.execute(sql, args)
            if is_query:
                data = cursor.fetchall()
            else:
                data = None
        except Exception as e:
            print(str(e))
            return 0, None
        else:
            if not is_query:
                conn.commit()
        finally:
            if cursor is not None:
                cursor.close()
            conn.close()

        return rows, data

    def _table_exists(self) -> bool:
        rows, _ = self._execute(self.SQL['table_exists'].format(
            db_name=self.pool_option['database'],
            table_name=self.table_name
        ))
        return rows > 0

    def _create_table(self):
        self._execute(self.SQL['create_table'].format(table_name=self.table_name))

    def load(self) -> List[ConfItem]:
        rows, data = self._execute(self.SQL['load_all'].format(table_name=self.table_name))
        return [
            self.row2item(data_item)
            for data_item in data
        ]

    def get(self, id_: str, default=None, value_only=True) -> Union[ConfItem, Any]:
        rows, data = self._execute(self.SQL['get_by_id'].format(table_name=self.table_name), id_)

        if rows:
            item = self.row2item(data[0])
            return item.value if value_only else item
        if id_ != self.VERSION_KEY:
            logger.warning('Conf item "%s" not found' % id_, stack_info=self.STACK_INFO)
        return default

    def match(self, prefix: str, value_only=True, fullkey=False) -> Dict[str, Union[ConfItem, Any]]:
        if prefix[-1] != '.':
            prefix = prefix + '.'
        prefix_len = len(prefix)

        rows, data = self._execute(self.SQL['match_items'].format(table_name=self.table_name), prefix + '%')

        items = {}
        for data_item in data:
            item = self.row2item(data_item)
            items[item.id if fullkey else item.id[prefix_len:]] = item.value if value_only else item

        return items

    def set(self, id_: str, value, update_sys_value=False, allow_add=False):
        rows, _ = self._execute(self.SQL['get_by_id'].format(table_name=self.table_name), id_)
        if not rows:
            if not allow_add:
                logger.warning('Conf item "%s" not found' % id_, stack_info=self.STACK_INFO)
                return
            self._add_item(id_, value, update_sys_value)
            return

        self._set_item(id_, value, update_sys_value)

    def batch_set(self, items: Union[Dict[str, Any], List[ConfItem]], prefix: str = None, update_sys_value=False,
                  allow_add=False):
        if prefix:
            if prefix[-1] != '.':
                prefix = prefix + '.'
        else:
            prefix = ''

        if isinstance(items, list):
            items = {
                item.id: item
                for item in items
            }

        all_items = self.load()
        data_map = {}

        for item in all_items:
            data_map[item.id] = item

        for key in items:
            id_ = prefix + key
            item = items[key]
            if key in data_map:
                self._set_item(id_, item, update_sys_value)
                continue
            if not allow_add:
                logger.warning('Conf item "%s" not found' % id_, stack_info=self.STACK_INFO)
                continue
            self._add_item(id_, item, update_sys_value)

    def remove(self, id_: Union[str, List[str], Tuple[str]]) -> Optional[Union[ConfItem, List[ConfItem]]]:
        is_list = isinstance(id_, (list, tuple))
        if is_list:
            id_list = id_
        else:
            id_list = (id_,)

        all_items = self.load()
        data_map = {}

        for item in all_items:
            data_map[item.id] = item

        remove_list = []

        for id_ in id_list:
            if id_ not in data_map:
                logger.warning('Conf item "%s" not found' % id_, stack_info=self.STACK_INFO)
                continue
            self._execute(self.SQL['remove_item'].format(table_name=self.table_name), id_)
            remove_list.append(data_map[id_])

        if is_list:
            return remove_list
        return None if len(remove_list) == 0 else remove_list[0]

    def clear(self):
        """
        清空配置表
        :return:
        """
        self._execute(self.SQL['clear_table'].format(table_name=self.table_name))

    def row2item(self, model: dict) -> ConfItem:
        value_type = model['value_type']
        id_ = model['id']

        item = ConfItem(self.parser)
        item.id = id_
        item.name = model['name']
        item.raw_sys_value = self._load_value(id_, value_type, model['value'])
        item.raw_user_value = self._load_value(id_, value_type, model['user_value'])
        item.type = ConfTypes(model['type'])
        item.desc = model['desc']
        item.enabled = model['enabled'] == b'\x01'

        return item

    def _set_item(self, id_, value, update_sys_value: bool):
        if isinstance(value, ConfItem):
            sys_value = value.raw_sys_value
            user_value = value.raw_user_value
            # 仅使用系统值的类型
            value_type = None if sys_value is None else type(sys_value).__name__
            sys_value = self._dump_value(sys_value)
            user_value = self._dump_value(user_value)
            self._execute(self.SQL['update_by_id'].format(table_name=self.table_name),
                          value.name, sys_value, user_value, value_type,
                          value.type.value, value.desc, value.enabled, id_)
        else:
            value_type = None if value is None else type(value).__name__
            value = self._dump_value(value)

            value_field = 'value' if update_sys_value else 'user_value'
            self._execute(self.SQL['update_value_by_id'].format(
                table_name=self.table_name, value_field=value_field
            ), value, value_type, id_)

    def _add_item(self, id_, value, update_sys_value):
        if isinstance(value, ConfItem):
            sys_value = value.raw_sys_value
            user_value = value.raw_user_value
            # 仅使用系统值的类型
            item = value
            item.value_type = None if sys_value is None else type(sys_value).__name__
            item.raw_sys_value = self._dump_value(sys_value)
            item.raw_user_value = self._dump_value(user_value)
        else:
            item = ConfItem()
            item.value_type = None if value is None else type(value).__name__
            item.update(value, update_sys_value)

        item.id = id_

        self._execute(self.SQL['add_item'].format(table_name=self.table_name),
                      item.id, item.name, item.raw_sys_value, item.raw_user_value, item.value_type,
                      item.type.value, item.desc, item.enabled)

    @classmethod
    def _load_value(cls, id_, value_type, value):
        if value_type is None or value_type == 'NoneType':
            return value

        if value is None:
            return value

        if value_type == 'int':
            return int(value)

        if value_type == 'bool':
            return value.lower() == 'true'

        if value_type == 'float':
            return float(value)

        if value_type in (dict, list):
            return json.loads(value)

        if value_type == 'str':
            return value

        if value_type in ('list', 'dict', 'tuple'):
            if value is None:
                return value

            return json.loads(value)

        logger.error('Cannot resolve value type "%s" of "%s"' % (value_type, id_))
        return value

    @classmethod
    def _dump_value(cls, value):
        if isinstance(value, bool):
            value = 'true' if value else 'false'
        elif isinstance(value, (dict, list, tuple)):
            value = None if value is None else json.dumps(value, ensure_ascii=False)

        return value



import json
import os
from typing import Any, Dict, List, Union, Tuple, Optional

from .base import ConfItem, IConf, logger
from .parser import ValueParser


class ConfFS(IConf):
    """
    基于文件系统的配置存储
    此模块基于文件进行缓存，缓存的文件绝对路径

    此模块会自动根据配置文件的修改时间来判断是否需要更新缓存

    _cached_items 存储了文件的最后修改时间以及文件内的配置项
    """

    def __init__(self, conf_root: str, parser: ValueParser = None):
        self.conf_root = os.path.abspath(conf_root)
        self._cached_items = {}
        logger.debug('Initialize filesystem configuration with path ' + self.conf_root)
        super(ConfFS, self).__init__(parser)

    def load(self, filename: str = None) -> List[ConfItem]:
        """

        :param filename:
        :return:
        """
        if filename is not None:
            # 文件如果不存在，返回空集合
            if not os.path.isfile(filename):
                return []

            # 获取此文件的最后修改时间
            mtime = os.path.getmtime(filename)
            # 从缓存查询此文件的配置项
            if filename in self._cached_items:
                [last_mtime, items] = self._cached_items[filename]
                if last_mtime == mtime:
                    return items
            # 最后修改时间不一致，那么更新
            items = self._load_conf_file(filename)
            self._cached_items[filename] = [mtime, items]
            return items

        items = []
        for filename in self._get_conf_files():
            items += self._load_conf_file(filename)

        return items

    def get(self, id_: str, default=None, value_only=True) -> Union[ConfItem, Any]:
        id_ = id_.lower()
        conf_prefix, conf_file = self._resolve_conf_file(id_)
        if conf_file is None:
            return default

        items = self.load(conf_file)

        for item in items:
            if item.id != id_:
                continue
            return item.value if value_only else item

        if id_ != self.VERSION_KEY:
            logger.warning('Conf item "%s" not found' % id_, stack_info=self.STACK_INFO)
        return default

    def match(self, prefix: str, value_only=True, fullkey=False) -> Dict[str, Union[ConfItem, Any]]:
        prefix = prefix.lower()
        if prefix[-1] != '.':
            prefix = prefix + '.'
        prefix_len = len(prefix)

        matched_files = self._resolve_matched_files(prefix)

        items = {}
        for filename in matched_files:
            for item in self.load(filename):
                if not item.id.startswith(prefix):
                    continue
                items[item.id if fullkey else item.id[prefix_len:]] = item.value if value_only else item
        return items

    def set(self, id_: str, value, update_sys_value=False, allow_add=False):
        id_ = id_.lower()
        conf_prefix, conf_file = self._resolve_conf_file(id_, True)

        found = False
        items = self.load(conf_file)
        for item in items:
            if item.id != id_:
                continue

            self._update_item(item, value, update_sys_value)
            found = True
            break

        if not found:
            if not allow_add:
                logger.warning('Conf item "%s" not found' % id_, stack_info=self.STACK_INFO)
                return
            item = self._add_item(id_, value, update_sys_value)
            items.append(item)

        # 将改动写入文件
        with open(conf_file, mode='w', encoding='utf8') as fp:
            json.dump(items, fp, indent=2, ensure_ascii=False, cls=ConfItem.JSONEncoder)

        if id_ != self.VERSION_KEY:
            self.version(True)

    def batch_set(self, items: Union[Dict[str, Any], List[ConfItem]], prefix: str = None, update_sys_value=False,
                  allow_add=False):
        if prefix:
            if prefix[-1] != '.':
                prefix = prefix + '.'
        else:
            prefix = ''

        if isinstance(items, list):
            items = {
                item.id: item
                for item in items
            }

        all_items = self.load()
        data_map = {}
        file_buffer = {}
        for item in all_items:
            data_map[item.id] = item
            pkg = '.'.join(item.id.split('.')[0:-1])
            if pkg not in file_buffer:
                conf_prefix, conf_file = self._resolve_conf_file(item.id, True)
                file_buffer[pkg] = {
                    'file': conf_file,
                    'changed': False,
                    'items': []
                }
            file_buffer[pkg]['items'].append(item)

        for key in items:
            id_ = prefix + key
            if id_ == self.VERSION_KEY:
                # ignore this item
                continue
            pkg = '.'.join(id_.split('.')[0:-1])
            item = items[key]
            if id_ in data_map:
                self._update_item(data_map[id_], item, update_sys_value)
            else:
                if not allow_add:
                    logger.warning('Conf item "%s" not found' % id_, stack_info=self.STACK_INFO)
                    continue
                if pkg not in file_buffer:
                    _, conf_file = self._resolve_conf_file(id_, True)
                    file_buffer[pkg] = {
                        'file': conf_file,
                        'changed': True,
                        'items': []
                    }
                item = self._add_item(id_, item, update_sys_value)
                file_buffer[pkg]['items'].append(item)

            file_buffer[pkg]['changed'] = True

        self._persist_changes(file_buffer)

    def remove(self, id_: Union[str, List[str], Tuple[str]]) -> Optional[Union[ConfItem, List[ConfItem]]]:
        is_list = isinstance(id_, (list, tuple))
        if is_list:
            id_list = id_
        else:
            id_list = (id_,)

        all_items = self.load()
        data_map = {}
        file_buffer = {}
        for item in all_items:
            data_map[item.id] = item
            pkg = '.'.join(item.id.split('.')[0:-1])
            if pkg not in file_buffer:
                conf_prefix, conf_file = self._resolve_conf_file(item.id, True)
                file_buffer[pkg] = {
                    'file': conf_file,
                    'changed': False,
                    'items': []
                }
            file_buffer[pkg]['items'].append(item)

        remove_list = []

        for id_ in id_list:
            if id_ == self.VERSION_KEY:
                # ignore this item
                continue
            if id_ not in data_map:
                logger.warning('Conf item "%s" not found' % id_, stack_info=self.STACK_INFO)
                continue
            pkg = '.'.join(id_.split('.')[0:-1])
            item = data_map[id_]
            file_buffer[pkg]['items'].remove(item)
            file_buffer[pkg]['changed'] = True
            remove_list.append(item)

        self._persist_changes(file_buffer)

        if is_list:
            return remove_list
        return None if len(remove_list) == 0 else remove_list[0]

    def _persist_changes(self, file_buffer: dict):
        has_changed = False
        for item in file_buffer.values():
            if not item['changed']:
                continue
            has_changed = True
            # 将改动写入文件
            with open(item['file'], mode='w', encoding='utf8') as fp:
                json.dump(item['items'], fp, indent=2, ensure_ascii=False, cls=ConfItem.JSONEncoder)

        if has_changed:
            self.version(True)

    @classmethod
    def _update_item(cls, item: ConfItem, value, update_sys_value):
        if isinstance(value, ConfItem):
            item.name = value.name
            item.raw_sys_value = value.raw_sys_value
            item.raw_user_value = value.raw_user_value
            item.raw_user_value = value.raw_user_value
            item.type = value.type
            item.desc = value.desc
        else:
            item.update(value, update_sys_value)

    @classmethod
    def _add_item(cls, id_, value, update_sys_value):
        if isinstance(value, ConfItem):
            item = value
        else:
            item = ConfItem()
            item.update(value, update_sys_value)

        item.id = id_

        return item

    def _resolve_conf_file(self, id_: str, auto_create=False):
        # 1. 去掉配置名称里面的最后一项(即得到配置文件路径)
        conf_dir = id_.split('.')[:-1]
        conf_prefix = '.'.join(conf_dir)
        # 2. 得到配置文件名称
        filename = conf_dir.pop() + '.json'

        # 3. 附加上配置文件目录（如果有）
        if len(conf_dir) > 0:
            conf_dir = os.path.join(self.conf_root, os.path.sep.join(conf_dir))
        else:
            conf_dir = self.conf_root

        fullname = os.path.join(conf_dir, filename)

        if os.path.isfile(fullname):
            return conf_prefix, fullname

        if id_ == self.VERSION_KEY or auto_create:
            if not os.path.isdir(conf_dir):
                os.makedirs(conf_dir, exist_ok=True)
            return conf_prefix, fullname

        logger.error('Cannot resolve id "%s": file "%s" not found' % (id_, fullname))
        return conf_prefix, None

    def _resolve_matched_files(self, key_prefix: str):
        files = []

        # 1. 去掉配置名称里面的 . 符号，得到路径
        path_prefix = key_prefix.rstrip('.').replace('.', os.path.pathsep)

        # 此时有两种可能性（可能同时存在）
        # 1. 配置前缀匹配到文件
        # 2. 配置前缀匹配到目录

        filename = os.path.join(self.conf_root, path_prefix + '.json')
        if os.path.isfile(filename):
            files.append(filename)

        dirname = os.path.join(self.conf_root, path_prefix)
        if os.path.isdir(dirname):
            for f in self._get_conf_files(dirname):
                files.append(f)

        if not files:
            logger.error('Cannot resolve id prefix "%s": path not found' % key_prefix)

        return files

    def _load_conf_file(self, filename: str) -> List[ConfItem]:
        with open(filename, mode='r', encoding='utf8') as fp:
            json_content = json.load(fp)

        pkg_prefix = os.path.splitext(os.path.relpath(filename, self.conf_root))[0].replace(os.path.sep, '.')

        items = []

        for json_item in json_content:
            item = ConfItem.decode(json_item, filename, self.parser)
            if not item:
                continue
            item.id = '%s.%s' % (pkg_prefix, item.id)
            items.append(item)

        return items

    def _get_conf_files(self, target_path: str = None):
        for current_dir, dirs, files in os.walk(target_path or self.conf_root):
            for filename in files:
                if filename.endswith('.json'):
                    yield os.path.join(current_dir, filename)



from enum import Enum
from logging import getLogger

logger = getLogger('confio')


class ConfTypes(Enum):
    """
    配置项的类型
    """

    NONE = ''
    """
    无类型
    """

    EXPR = 'expr'
    """
    表达式
    """

    PATH = 'path'
    """
    路径
    """

    SIZE = 'size'
    """
    文件大小
    """

    TIME = 'time'
    """
    时间值
    """


SIZE = {
    'b': 1,
    'k': 1024,
    'm': 1024 ** 2,
    'g': 1024 ** 3,
    't': 1024 ** 4,
    'p': 1024 ** 5
}

TIME = {
    's': 1,
    'm': 60,
    'h': 60 ** 2,
    'D': 60 ** 2 * 24,
    'M': 60 ** 2 * 24 * 30,
    'Y': 60 ** 2 * 24 * 365,
}


class ValueParser:
    PATH_ROOT = None

    def __init__(self, path_root: str = None):
        self.path_root = path_root or self.PATH_ROOT

    def parse_path(self, value):
        import os
        if not os.path.isabs(value):
            if self.path_root is None:
                raise Exception(
                    'The path root must be specified via `ValueParser.PATH_ROOT=` or `ValueParser(path_root=)`'
                )
            value = os.path.join(self.path_root, value)
        return os.path.abspath(value)

    @classmethod
    def parse_expr(cls, value):
        try:
            return eval(value)
        except Exception as e:
            logger.warning('Cannot evaluate expression "%s": %s' % (value, repr(e)))
            return value

    @classmethod
    def parse_size(cls, value):
        try:
            temp = []
            for ch in value:
                if not ch.isalpha():
                    temp.append(ch)
                    continue
                temp.append('*%s+' % SIZE[ch.lower()])
            # 最后的右边可能存在一个多余的 + 符号
            expr = ''.join(temp).rstrip('+')
            return eval(expr)
        except Exception as e:
            logger.warning('Cannot evaluate size "%s": %s' % (value, repr(e)))
            return int(value)

    @classmethod
    def parse_time(cls, value):
        try:
            temp = []
            for ch in value:
                if not ch.isalpha():
                    temp.append(ch)
                    continue
                temp.append('*%s+' % TIME[ch])
            # 最后的右边可能存在一个多余的 + 符号
            expr = ''.join(temp).rstrip('+')
            return eval(expr)
        except Exception as e:
            logger.warning('Cannot evaluate time "%s": %s' % (value, repr(e)))
            return int(value)



import re

import setuptools

with open("src/confio/__meta__.py", encoding="utf8") as f:
    version = re.search(r'version = ([\'"])(.*?)\1', f.read()).group(2)

setuptools.setup(
    name='confio',
    version=version,
    install_requires=[]
)



from .base import ConfItem, IConf
from .db import ConfDB
from .fs import ConfFS
from .parser import ValueParser, TIME, SIZE, ConfTypes

__all__ = [
    'ConfItem',
    'IConf',
    'ConfFS',
    'ConfDB',
    'TIME',
    'SIZE',
    'ConfTypes',
    'ValueParser'
]



import json
from abc import ABC, abstractmethod
from logging import getLogger
from typing import Any, Dict, List, Union, Tuple, Optional

from .parser import ConfTypes, ValueParser
from .__meta__ import name

logger = getLogger('confio')


class ConfItem:
    """
    配置项结构定义
    """

    def __init__(self, parser: ValueParser = None):
        self.id = None
        self.name = None
        self.raw_sys_value = None
        self.raw_user_value = None
        self.desc = None
        self.enabled = True
        self.type = ConfTypes.NONE
        self.parser = parser or ValueParser()

    def update(self, value, update_sys_value=False):
        """
        更新值
        :param value: 新的值
        :param update_sys_value: 为 True 时表示更新系统值，为 False 表示更新用户值
        :return:
        """
        if update_sys_value:
            self.raw_sys_value = value
        else:
            self.raw_user_value = value

    def __str__(self):
        return '<ConfItem %s:%s=%s -%s>' % (self.id, self.type.name, self.value, self.name)

    def _parse_value(self, raw_value):
        """
        将原始值解析出来
        :param raw_value:
        :return:
        """
        if raw_value is None:
            return None

        if self.type == ConfTypes.NONE:
            return raw_value

        return getattr(self.parser, 'parse_%s' % self.type.value)(raw_value)

    @property
    def sys_value(self):
        """
        系统值
        :return:
        """
        return self._parse_value(self.raw_sys_value)

    @property
    def user_value(self):
        """
        用户值
        :return:
        """
        return self._parse_value(self.raw_user_value)

    @property
    def value(self):
        """
        获取用户值或者系统值（用户值未指定：为 None 时，返回系统值）
        :return:
        """
        if self.raw_user_value is None:
            return self.sys_value
        return self.user_value

    class JSONEncoder(json.JSONEncoder):
        def default(self, o):
            if not isinstance(o, ConfItem):
                return json.JSONEncoder().default(o)

            result = {
                'name': o.name,
                'key': o.id.split('.')[-1]
            }

            if o.type == ConfTypes.NONE:
                result['value'] = o.raw_sys_value
            else:
                result['value:%s' % o.type.value] = o.raw_sys_value

            if o.raw_user_value is not None:
                result['user_value'] = o.raw_user_value

            if o.desc:
                result['desc'] = o.desc

            # 仅在 false 时才写此值输出
            # 默认值即为 True
            if o.enabled is False:
                result['enabled'] = o.enabled

            return result

    @classmethod
    def decode(cls, json_data: dict, filename: str, parser: ValueParser = None):
        """
        从 JSON 解析数据，生成 ConfItem 对象
        :param json_data:
        :param filename:
        :param parser:
        :return:
        """
        if not isinstance(json_data, dict):
            return None

        key = cls._get_value(json_data, 'key')

        if not key or not isinstance(key, str) or '.' in key:
            raise Exception('Invalid config key in file "%s": %s' % (filename, key))

        if not key.islower():
            logger.debug('Conf key should be lowercase: %s' % key)
            key = key.lower()

        item = ConfItem(parser)
        item.id = key
        item.raw_user_value = cls._get_value(json_data, 'user_value')
        item.name = cls._get_value(json_data, 'name')
        item.desc = cls._get_value(json_data, 'desc')
        item.enabled = cls._get_value(json_data, 'enabled', True)

        value, item_type = cls._get_value_and_type(json_data)

        item.raw_sys_value = value
        item.type = item_type

        return item

    @classmethod
    def _get_value(cls, source: dict, field: str, default=None):
        return source[field] if field in source else default

    @classmethod
    def _get_value_and_type(cls, json_data: dict):
        value = None
        type_name = None
        for field in json_data:
            temp = field.split(':')
            if len(temp) == 1:
                continue
            type_name = temp[1]
            value = json_data[field]
            break

        if type_name is None:
            return cls._get_value(json_data, 'value'), ConfTypes.NONE

        try:
            item_type = ConfTypes(type_name)
        except Exception:
            logger.error('Cannot cast value %r into type "ConfTypes"' % type_name)
            item_type = ConfTypes.NONE

        return value, item_type


class IConf(ABC):
    STACK_INFO = False
    """
    在找不到配置项时，是否显示堆栈信息
    """

    VERSION_KEY = '%s.options.version' % name
    """
    版本号标识项名称
    """

    def __init__(self, parser: ValueParser = None):
        self.parser = parser

    def exists(self, id_: str) -> bool:
        """
        检查指定的配置项是否存在
        """
        return self.get(id_, value_only=False) is not None

    @abstractmethod
    def load(self, **kwargs) -> List[ConfItem]:
        """
        加载所有配置项（列表）
        """
        pass

    @abstractmethod
    def get(self, id_: str, default=None, value_only=True) -> Union[ConfItem, Any]:
        """
        获取指定的配置项的值，当配置项不存在时，使用指定的默认值
        """
        pass

    @abstractmethod
    def match(self, prefix: str, value_only=True, fullkey=False) -> Dict[str, Union[ConfItem, Any]]:
        """
        匹配指定前缀的项，返回的每个配置
        """
        pass

    @abstractmethod
    def set(self, id_: str, value, update_sys_value=False, allow_add=False):
        """
        设置指定配置项的值
        :param id_: 需要更新/添加项的键名称
        :param value: 需要更新/添加项的值
        :param update_sys_value: 为 True 时表示更新系统值，为 False 表示更新用户值
        :param allow_add: 在配置项不存在时，是否允许添加
        :return:
        """
        pass

    @abstractmethod
    def batch_set(self, items: Union[Dict[str, Any], List[ConfItem]], prefix: str = None, update_sys_value=False,
                  allow_add=False):
        """
        批量设置配置项
        :param items: 要更新的配置项集合
        :param prefix: 前缀，会被添加到 items 的 key 的前面，不需要填写 `.` 符号
        :param update_sys_value: 是否更新 sys_value ，默认更新 user_value
        :param allow_add: 在配置项不存在时，是否允许添加
        :return:
        """
        pass

    @abstractmethod
    def remove(self, id_: Union[str, List[str], Tuple[str]]) -> Optional[Union[ConfItem, List[ConfItem]]]:
        """
        根据配置项的ID移除一个或多个
        :param id_:
        :return:
        """
        pass

    def version(self, update=False):
        old = self.get(self.VERSION_KEY, value_only=False)
        if old is None:
            old = ConfItem()
            old.name = 'Confit options version identifier'
            old.raw_sys_value = 0
            self.set(self.VERSION_KEY, old, allow_add=True)

        if update:
            old.raw_sys_value = old.raw_sys_value + 1
            self.set(self.VERSION_KEY, old.raw_sys_value, update_sys_value=True)

        return old.value

    def export(self, filename: str, title: str = None):
        """
        导出为 markdown 格式文件
        :param title:
        :param filename:
        :return:
        """
        items = self.load()
        items.sort(key=lambda item: item.id)

        from .__meta__ import website, name

        headings = []

        with open(filename, mode='w', encoding='utf8') as fp:
            fp.write('# %s\n\n此文档由 [%s](%s) 生成\n' % (
                title or '配置项说明',
                name,
                website
            ))
            for item in items:
                if item.id == self.VERSION_KEY:
                    continue
                temp = item.id.split('.')
                heading = temp[0]
                if heading not in headings:
                    headings.append(heading)
                    fp.write('\n## %s\n\n' % heading)
                fp.write('- `{id}` {name}\n\t> 描述: {desc}\n'.format(
                    id=item.id,
                    name=item.name or '_未命名_',
                    desc=item.desc or '_无_')
                )
